@@ -24,6 +24,7 @@
 #endif
 #include "gtkimhtml.h"
 #include <gtk/gtk.h>
+#include <gdk/gdkkeysyms.h>
 #include <string.h>
 #include <ctype.h>
 #include <stdio.h>
@@ -33,6 +34,19 @@
 #include <locale.h>
 #endif
 
+#ifdef ENABLE_NLS
+#  include <libintl.h>
+#  define _(x) gettext(x)
+#  ifdef gettext_noop
+#    define N_(String) gettext_noop (String)
+#  else
+#    define N_(String) (String)
+#  endif
+#else
+#  define N_(String) (String)
+#  define _(x) (x)
+#endif
+
 
 /* POINT_SIZE converts from AIM font sizes to point sizes.  It probably should be redone in such a
  * way that it base the sizes off the default font size rather than using arbitrary font sizes. */
@@ -53,7 +67,7 @@ struct _FontDetail {
 struct _GtkSmileyTree {
 	GString *values;
 	GtkSmileyTree **children;
-	gchar *image;
+	GtkIMHtmlSmiley *image;
 };
 
 static GtkSmileyTree*
@@ -64,11 +78,10 @@ gtk_smiley_tree_new ()
 
 static void
 gtk_smiley_tree_insert (GtkSmileyTree *tree,
-			const gchar   *text,
-			const gchar   *path)
+			GtkIMHtmlSmiley *smiley)
 {
 	GtkSmileyTree *t = tree;
-	const gchar *x = text;
+	const gchar *x = smiley->smile;
 
 	if (!strlen (x))
 		return;
@@ -94,9 +107,11 @@ gtk_smiley_tree_insert (GtkSmileyTree *tree,
 		x++;
 	}
 	
-	t->image = path;
+	t->image = smiley;
 }
-gtk_smiley_tree_destroy (GtkSmileyTree *tree)
+
+
+void gtk_smiley_tree_destroy (GtkSmileyTree *tree)
 {
 	GSList *list = g_slist_append (NULL, tree);
 
@@ -114,6 +129,7 @@ gtk_smiley_tree_destroy (GtkSmileyTree *tree)
 	}
 }
 
+
 static GtkTextViewClass *parent_class = NULL;
 
 
@@ -124,13 +140,23 @@ enum {
 };
 static guint signals [LAST_SIGNAL] = { 0 };
 
+static gboolean
+gtk_smiley_tree_destroy_from_hash(gpointer key, gpointer value,
+		gpointer user_data)
+{
+	gtk_smiley_tree_destroy(value);
+	return TRUE;
+}
+
 static void
 gtk_imhtml_finalize (GObject *object)
 {
 	GtkIMHtml *imhtml = GTK_IMHTML(object);
-
-	g_hash_table_foreach_remove(imhtml->smiley_data, gtk_smiley_tree_destroy, NULL);
+	g_hash_table_foreach_remove(imhtml->smiley_data, gtk_smiley_tree_destroy_from_hash, NULL);
+	g_hash_table_destroy(imhtml->smiley_data);
 	gtk_smiley_tree_destroy(imhtml->default_smilies);
+	gdk_cursor_unref(imhtml->hand_cursor);
+	gdk_cursor_unref(imhtml->arrow_cursor);
 	G_OBJECT_CLASS(parent_class)->finalize (object);
 }
 
@@ -142,13 +168,15 @@ static void gtk_imhtml_class_init (GtkIMHtmlClass *class)
 	object_class = (GtkObjectClass*) class;
 	gobject_class = (GObjectClass*) class;
 	parent_class = gtk_type_class(GTK_TYPE_TEXT_VIEW);
-	signals[URL_CLICKED] = gtk_signal_new("url_clicked",
-					      GTK_RUN_FIRST,
-					      GTK_CLASS_TYPE(object_class),
-					      GTK_SIGNAL_OFFSET(GtkIMHtmlClass, url_clicked),
-					      gtk_marshal_NONE__POINTER,
-					      GTK_TYPE_NONE, 1,
-					      GTK_TYPE_POINTER);
+	signals[URL_CLICKED] = g_signal_new("url_clicked",
+						G_TYPE_FROM_CLASS(gobject_class),
+						G_SIGNAL_RUN_FIRST,
+						G_STRUCT_OFFSET(GtkIMHtmlClass, url_clicked),
+						NULL,
+						0,
+						g_cclosure_marshal_VOID__POINTER,
+						G_TYPE_NONE, 1,
+						G_TYPE_POINTER);
 	gobject_class->finalize = gtk_imhtml_finalize;
 }
 
@@ -179,6 +207,9 @@ static void gtk_imhtml_init (GtkIMHtml *imhtml)
 	imhtml->hand_cursor = gdk_cursor_new (GDK_HAND2);
 	imhtml->arrow_cursor = gdk_cursor_new (GDK_LEFT_PTR);
 
+	imhtml->show_smileys = TRUE;
+	imhtml->show_comments = TRUE;
+
 	imhtml->smiley_data = g_hash_table_new (g_str_hash, g_str_equal);
 	imhtml->default_smilies = gtk_smiley_tree_new();
 }
@@ -209,25 +240,105 @@ GtkType gtk_imhtml_get_type()
 	return imhtml_type;
 }
 
-/* The call back for an event on a link tag. */
-void tag_event(GtkTextTag *tag, GObject *arg1, GdkEvent *event, GtkTextIter *arg2, char *url) {
+struct url_data {
+	GObject *object;
+	gchar *url;
+};
+
+static void url_open(GtkWidget *w, struct url_data *data) {
+	if(!data) return;
+
+	g_signal_emit(data->object, signals[URL_CLICKED], 0, data->url);
+	
+	g_object_unref(data->object);
+	g_free(data->url);
+	g_free(data);
+}
+static void url_copy(GtkWidget *w, gchar *url) {
+	GtkClipboard *clipboard;
+
+	clipboard = gtk_clipboard_get(GDK_SELECTION_CLIPBOARD);
+	gtk_clipboard_set_text(clipboard, url, -1);
+}
+
+/* The callback for an event on a link tag. */
+gboolean tag_event(GtkTextTag *tag, GObject *arg1, GdkEvent *event, GtkTextIter *arg2, char *url) {
+	GdkEventButton *event_button = (GdkEventButton *) event;
+
 	if (event->type == GDK_BUTTON_RELEASE) {
-		/* A link was clicked--we emit the "url_clicked" signal with the URL as the argument */
-		//	if ((GdkEventButton)(event)->button == 1) 
-			gtk_signal_emit (G_OBJECT(arg1), signals[URL_CLICKED], url);
+		if (event_button->button == 1) { 
+			GtkTextIter start, end;
+			/* we shouldn't open a URL if the user has selected something: */
+			gtk_text_buffer_get_selection_bounds(
+						gtk_text_iter_get_buffer(arg2),	&start, &end);
+			if(gtk_text_iter_get_offset(&start) !=
+					gtk_text_iter_get_offset(&end))
+				return FALSE;
+
+			/* A link was clicked--we emit the "url_clicked" signal
+			 * with the URL as the argument */
+			g_signal_emit(arg1, signals[URL_CLICKED], 0, url);
+			return FALSE;
+		} else if(event_button->button == 3) {
+			GtkWidget *img, *item, *label, *menu;
+			struct url_data *tempdata = g_new(struct url_data, 1);
+			tempdata->object = g_object_ref(arg1);
+			tempdata->url = g_strdup(url);
+		
+			menu = gtk_menu_new();
+			gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, 
+							event_button->button, event_button->time);
+			
+			/* URL of link, desensitized */
+			item = gtk_menu_item_new();
+			label = gtk_label_new(url);
+			gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
+			gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
+			gtk_container_add(GTK_CONTAINER(item), label);
+			gtk_widget_set_sensitive(item, FALSE);
+			gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
+
+			/* seperator */
+			item = gtk_menu_item_new();
+			gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
+						
+			/* buttons and such */
+			img = gtk_image_new_from_stock(GTK_STOCK_COPY, GTK_ICON_SIZE_MENU);
+			item = gtk_image_menu_item_new_with_mnemonic(_("_Copy Link Location"));
+			gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(item), img);
+			g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(url_copy),
+					url);
+			gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
+
+			img = gtk_image_new_from_stock(GTK_STOCK_JUMP_TO, GTK_ICON_SIZE_MENU);
+			item = gtk_image_menu_item_new_with_mnemonic(_("_Open Link in Browser"));
+			gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(item), img);
+			g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(url_open),
+					tempdata);
+			gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
+			gtk_widget_show_all(menu);
+			
+			return TRUE;
+		}
 	} else if (event->type == GDK_ENTER_NOTIFY) {
 		/* make a hand cursor and a tooltip timeout -- if GTK worked as it should */
 	} else if (event->type == GDK_LEAVE_NOTIFY) {
 		/* clear timeout and make an arrow cursor again --if GTK worked as it should */
 	}
+	if(event->type == GDK_BUTTON_PRESS && event_button->button == 3)
+		return TRUE; /* Clicking the right mouse button on a link shouldn't
+						be caught by the regular GtkTextView menu */
+	else
+		return FALSE; /* Let clicks go through if we didn't catch anything */
 }
 
+/* this isn't used yet
 static void
-gtk_smiley_tree_remove (GtkSmileyTree *tree,
-			const gchar   *text)
+gtk_smiley_tree_remove (GtkSmileyTree     *tree,
+			GtkIMHtmlSmiley   *smiley)
 {
 	GtkSmileyTree *t = tree;
-	const gchar *x = text;
+	const gchar *x = smiley->smile;
 	gint len = 0;
 
 	while (*x) {
@@ -245,9 +356,12 @@ gtk_smiley_tree_remove (GtkSmileyTree *tree,
 		x++; len++;
 	}
 
-	if (t->image)
+	if (t->image) {
 		t->image = NULL;
+	}
 }
+*/
+
 
 static gint
 gtk_smiley_tree_lookup (GtkSmileyTree *tree,
@@ -279,16 +393,14 @@ gtk_smiley_tree_lookup (GtkSmileyTree *tree,
 }
 
 void
-gtk_imhtml_associate_smiley (GtkIMHtml  *imhtml,
-			     gchar      *text,
-			     gchar      *sml,
-			     gchar      *path)
+gtk_imhtml_associate_smiley (GtkIMHtml       *imhtml,
+			     gchar           *sml,
+			     GtkIMHtmlSmiley *smiley)
 {
 	GtkSmileyTree *tree;
 	g_return_if_fail (imhtml != NULL);
 	g_return_if_fail (GTK_IS_IMHTML (imhtml));
-	g_return_if_fail (text != NULL);
-
+	
 	if (sml == NULL)
 		tree = imhtml->default_smilies;
 	else if ((tree = g_hash_table_lookup(imhtml->smiley_data, sml))) {
@@ -297,10 +409,7 @@ gtk_imhtml_associate_smiley (GtkIMHtml  *imhtml,
 		g_hash_table_insert(imhtml->smiley_data, sml, tree);
 	}
 
-	if (path == NULL)
-		gtk_smiley_tree_remove (tree, text);
-	else
-		gtk_smiley_tree_insert (tree, text, path);
+	gtk_smiley_tree_insert (tree, smiley);
 }
 
 static gboolean
@@ -330,14 +439,13 @@ gtk_imhtml_is_smiley (GtkIMHtml   *imhtml,
 	return (*len > 0);
 }
 
-static gchar*
+GdkPixbuf*
 gtk_smiley_tree_image (GtkIMHtml     *imhtml,
 		       const gchar   *sml,
 		       const gchar   *text)
 {
 	GtkSmileyTree *t;
 	const gchar *x = text;
-
 	if (sml == NULL)
 		t = imhtml->default_smilies;
 	else 
@@ -363,7 +471,10 @@ gtk_smiley_tree_image (GtkIMHtml     *imhtml,
 		x++;
 	}
 
-	return t->image;
+	if (!t->image->icon)
+		t->image->icon = gdk_pixbuf_new_from_file(t->image->file, NULL);
+
+	return t->image->icon;
 }
 #define VALID_TAG(x)	if (!g_strncasecmp (string, x ">", strlen (x ">"))) {	\
 				*tag = g_strndup (string, strlen (x));		\
@@ -561,6 +672,7 @@ gtk_imhtml_get_html_opt (gchar       *tag,
 
 #define NEW_TEXT_BIT 0
 #define NEW_HR_BIT 1
+#define NEW_COMMENT_BIT 2
 #define NEW_BIT(x)	ws [wpos] = '\0'; \
                         mark2 = gtk_text_buffer_create_mark(imhtml->text_buffer, NULL, &iter, TRUE); \
                         gtk_text_buffer_insert(imhtml->text_buffer, &iter, ws, -1); \
@@ -612,16 +724,15 @@ gtk_imhtml_get_html_opt (gchar       *tag,
                         wpos = 0; \
                         ws[0] = 0; \
                         gtk_text_buffer_get_end_iter(imhtml->text_buffer, &iter); \
-
-
-/*                        if (x == NEW_HR_BIT) { \
-	                         sep = gtk_hseparator_new(); \
-                                 gtk_widget_size_request(GTK_WIDGET(imhtml), &req); \
-                                 gtk_widget_set_size_request(sep, 20, -1); \
-	                         anchor = gtk_text_buffer_create_child_anchor(imhtml->text_buffer, &iter); \
+                        if (x == NEW_HR_BIT) { \
+                                 GtkTextChildAnchor *anchor = gtk_text_buffer_create_child_anchor(imhtml->text_buffer, &iter); \
+                                 GtkWidget *sep = gtk_hseparator_new(); \
+                                 GdkRectangle rect; \
+                                 gtk_text_view_get_visible_rect(GTK_TEXT_VIEW(imhtml), &rect); \
+                                 gtk_widget_set_size_request(GTK_WIDGET(sep), rect.width, 2); \
                                  gtk_text_view_add_child_at_anchor(GTK_TEXT_VIEW(imhtml), sep, anchor); \
                                  gtk_widget_show(sep); \
-			  } */
+                        } \
 
 GString* gtk_imhtml_append_text (GtkIMHtml        *imhtml,
 				 const gchar      *text,
@@ -632,10 +743,7 @@ GString* gtk_imhtml_append_text (GtkIMHtml        *imhtml,
 	GString *str = NULL;
 	GtkTextIter iter, siter;
 	GtkTextMark *mark, *mark2;
-	GtkTextChildAnchor *anchor;
 	GtkTextTag *texttag;
-	GtkWidget *sep;
-	GtkRequisition req;
 	gchar *ws;
 	gchar *tag;
 	gchar *url = NULL;
@@ -656,6 +764,9 @@ GString* gtk_imhtml_append_text (GtkIMHtml        *imhtml,
 
 	GSList *fonts = NULL;
 
+	GdkRectangle rect;
+	int y, height;
+
 	g_return_val_if_fail (imhtml != NULL, NULL);
 	g_return_val_if_fail (GTK_IS_IMHTML (imhtml), NULL);
 	g_return_val_if_fail (text != NULL, NULL);
@@ -672,6 +783,15 @@ GString* gtk_imhtml_append_text (GtkIMHtml        *imhtml,
 
 	gtk_text_buffer_get_end_iter(imhtml->text_buffer, &iter);
 	mark = gtk_text_buffer_create_mark (imhtml->text_buffer, NULL, &iter, /* right grav */ FALSE);
+
+	gtk_text_view_get_visible_rect(GTK_TEXT_VIEW(imhtml), &rect);	
+	gtk_text_view_get_line_yrange(GTK_TEXT_VIEW(imhtml), &iter, &y, &height);
+
+	if(((y + height) - (rect.y + rect.height)) > height 
+	   && gtk_text_buffer_get_char_count(imhtml->text_buffer)){
+		options |= GTK_IMHTML_NO_SCROLL;
+	}
+
 	while (pos < len) {
 		if (*c == '<' && gtk_imhtml_is_tag (c + 1, &tag, &tlen, &type)) {
 			c++;
@@ -771,6 +891,7 @@ GString* gtk_imhtml_append_text (GtkIMHtml        *imhtml,
 				case 42:        /* HR (opt) */
 					ws[wpos++] = '\n';
 					NEW_BIT(NEW_HR_BIT);
+					ws[wpos++] = '\n';
 					break;
 				case 27:	/* /FONT */
 					if (fonts) {
@@ -891,7 +1012,8 @@ GString* gtk_imhtml_append_text (GtkIMHtml        *imhtml,
 					break;
 				case 49:	/* comment */
 					NEW_BIT (NEW_TEXT_BIT);
-					wpos = g_snprintf (ws, len, "%s", tag);
+					if (imhtml->show_comments)
+						wpos = g_snprintf (ws, len, "%s", tag);
 					NEW_BIT (NEW_COMMENT_BIT);
 					break;
 				default:
@@ -899,6 +1021,8 @@ GString* gtk_imhtml_append_text (GtkIMHtml        *imhtml,
 				}
 			c += tlen;
 			pos += tlen;
+			if(tag)
+				g_free(tag); /* This was allocated back in VALID_TAG() */
 		} else if (*c == '&' && gtk_imhtml_is_amp_escape (c, &amp, &tlen)) {
 			ws [wpos++] = amp;
 			c += tlen;
@@ -911,7 +1035,7 @@ GString* gtk_imhtml_append_text (GtkIMHtml        *imhtml,
 			}
 			c++;
 			pos++;
-		} else if (gtk_imhtml_is_smiley (imhtml, fonts, c, &smilelen) || gtk_imhtml_is_smiley(imhtml, NULL, c, &smilelen)) {
+		} else if (imhtml->show_smileys && (gtk_imhtml_is_smiley (imhtml, fonts, c, &smilelen) || gtk_imhtml_is_smiley(imhtml, NULL, c, &smilelen))) {
 			FontDetail *fd;
 			gchar *sml = NULL;
 			if (fonts) {
@@ -920,7 +1044,7 @@ GString* gtk_imhtml_append_text (GtkIMHtml        *imhtml,
 			}
 			NEW_BIT (NEW_TEXT_BIT);
 			wpos = g_snprintf (ws, smilelen + 1, "%s", c);
-			gtk_text_buffer_insert_pixbuf(imhtml->text_buffer, &iter, gdk_pixbuf_new_from_file(gtk_smiley_tree_image (imhtml, sml, ws), NULL));
+			gtk_text_buffer_insert_pixbuf(imhtml->text_buffer, &iter, gtk_smiley_tree_image (imhtml, sml, ws));
 			c += smilelen;
 			pos += smilelen;
 			wpos = 0;
@@ -998,26 +1122,24 @@ GString* gtk_imhtml_append_text (GtkIMHtml        *imhtml,
 	return str;
 }
 
-void       gtk_imhtml_set_adjustments  (GtkIMHtml        *imhtml,
-					GtkAdjustment    *hadj,
-	GtkAdjustment    *vadj){}
-
-void       gtk_imhtml_set_defaults     (GtkIMHtml            *imhtml,
-					PangoFontDescription *font,
-					GdkColor             *fg_color,
-	GdkColor             *bg_color){}
-
-void       gtk_imhtml_set_img_handler  (GtkIMHtml        *imhtml,
-	GtkIMHtmlImage    handler){}
-
-void 	   gtk_imhtml_init_smileys     (GtkIMHtml *imhtml){}
-void       gtk_imhtml_remove_smileys   (GtkIMHtml        *imhtml){}
-void       gtk_imhtml_reset_smileys   (GtkIMHtml        *imhtml){}
+void gtk_imhtml_remove_smileys(GtkIMHtml *imhtml) 
+{ 
+	g_hash_table_destroy(imhtml->smiley_data);
+	gtk_smiley_tree_destroy(imhtml->default_smilies);
+	imhtml->smiley_data = g_hash_table_new (g_str_hash, g_str_equal);
+	imhtml->default_smilies = gtk_smiley_tree_new();
+}
 void       gtk_imhtml_show_smileys     (GtkIMHtml        *imhtml,
-	gboolean          show){}
+					gboolean          show)
+{
+	imhtml->show_smileys = show;
+}
 
 void       gtk_imhtml_show_comments    (GtkIMHtml        *imhtml,
-	gboolean          show){}
+					gboolean          show)
+{
+	imhtml->show_comments = show;
+}
 
 void
 gtk_imhtml_clear (GtkIMHtml *imhtml)
@@ -1029,5 +1151,8 @@ gtk_imhtml_clear (GtkIMHtml *imhtml)
 	gtk_text_buffer_delete(imhtml->text_buffer, &start, &end);
 }
 
-void       gtk_imhtml_page_up          (GtkIMHtml        *imhtml){}
+void gtk_imhtml_page_up (GtkIMHtml *imhtml)
+{
+
+}
 void       gtk_imhtml_page_down        (GtkIMHtml        *imhtml){}
