@@ -1,6 +1,6 @@
 /*
  * gaim - Gadu-Gadu Protocol Plugin
- * $Id: gg.c,v 1.37 2003/01/28 21:42:13 seanegan Exp $
+ * $Id: gg.c,v 1.44 2003/02/24 00:29:52 faceprint Exp $
  *
  * Copyright (C) 2001 Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
  * 
@@ -44,10 +44,10 @@
 #include <ctype.h>
 /* Library from EKG (Eksperymentalny Klient Gadu-Gadu) */
 #include "libgg.h"
+#include "gaim.h"
 #include "multi.h"
 #include "core.h"
 #include "prpl.h"
-#include "gaim.h"
 #include "proxy.h"
 
 #ifdef _WIN32
@@ -130,7 +130,7 @@ static gint args_compare(gconstpointer a, gconstpointer b)
 
 static gboolean allowed_uin(struct gaim_connection *gc, char *uin)
 {
-	switch (gc->user->permdeny) {
+	switch (gc->account->permdeny) {
 	case 1:
 		/* permit all, deny none */
 		return TRUE;
@@ -141,13 +141,13 @@ static gboolean allowed_uin(struct gaim_connection *gc, char *uin)
 		break;
 	case 3:
 		/* permit some. */
-		if (g_slist_find_custom(gc->user->permit, uin, args_compare))
+		if (g_slist_find_custom(gc->account->permit, uin, args_compare))
 			return TRUE;
 		return FALSE;
 		break;
 	case 4:
 		/* deny some. */
-		if (g_slist_find_custom(gc->user->deny, uin, args_compare))
+		if (g_slist_find_custom(gc->account->deny, uin, args_compare))
 			return FALSE;
 		return TRUE;
 		break;
@@ -278,7 +278,7 @@ static GList *agg_buddy_menu(struct gaim_connection *gc, char *who)
 {
 	GList *m = NULL;
 	struct proto_buddy_menu *pbm;
-	struct buddy *b = find_buddy(gc->user, who);
+	struct buddy *b = find_buddy(gc->account, who);
 	static char buf[AGG_BUF_LEN];
 
 	if (!b) {
@@ -306,7 +306,7 @@ static void main_callback(gpointer data, gint source, GaimInputCondition cond)
 	if (gd->sess->fd != source)
 		gd->sess->fd = source;
 
-	if (source != 0) {
+	if (source == 0) {
 		hide_login_progress(gc, _("Could not connect"));
 		signoff(gc);
 		return;
@@ -423,7 +423,7 @@ void login_callback(gpointer data, gint source, GaimInputCondition cond)
 	}
 	debug_printf("Found GG connection\n");
 	
-	if (source != 0) {
+	if (source == 0) {
 		hide_login_progress(gc, _("Unable to connect."));
 		signoff(gc);
 		return;
@@ -532,9 +532,9 @@ static void agg_keepalive(struct gaim_connection *gc)
 	}
 }
 
-static void agg_login(struct aim_user *user)
+static void agg_login(struct gaim_account *account)
 {
-	struct gaim_connection *gc = new_gaim_conn(user);
+	struct gaim_connection *gc = new_gaim_conn(account);
 	struct agg_data *gd = gc->proto_data = g_new0(struct agg_data, 1);
 	char buf[80];
 
@@ -542,13 +542,13 @@ static void agg_login(struct aim_user *user)
 
 	gd->sess = g_new0(struct gg_session, 1);
 
-	if (user->proto_opt[USEROPT_NICK][0])
+	if (account->proto_opt[USEROPT_NICK][0])
 		g_snprintf(gc->displayname, sizeof(gc->displayname), "%s",
-			   user->proto_opt[USEROPT_NICK]);
+			   account->proto_opt[USEROPT_NICK]);
 
 	set_login_progress(gc, 1, _("Looking up GG server"));
 
-	if (invalid_uin(user->username)) {
+	if (invalid_uin(account->username)) {
 		hide_login_progress(gc, _("Invalid Gadu-Gadu UIN specified"));
 		signoff(gc);
 		return;
@@ -567,8 +567,8 @@ static void agg_login(struct aim_user *user)
 	   gg_login() sucks for me, so I'm using proxy_connect()
 	 */
 
-	gd->sess->uin = (uin_t) strtol(user->username, (char **)NULL, 10);
-	gd->sess->password = g_strdup(user->password);
+	gd->sess->uin = (uin_t) strtol(account->username, (char **)NULL, 10);
+	gd->sess->password = g_strdup(account->password);
 	gd->sess->state = GG_STATE_CONNECTING;
 	gd->sess->check = GG_CHECK_WRITE;
 	gd->sess->async = 1;
@@ -691,7 +691,7 @@ static void search_results(struct gaim_connection *gc, gchar *webdata)
 			ibuf = g_strconcat("---------------------------------<BR>\n", NULL);
 			oldibuf = ibuf;
 			ibuf = g_strconcat(oldibuf, "<B>", _("Active"), ":</B> ",
-					   (atoi(p) == 2) ? _("yes") : _("no"), "<BR>\n", NULL);
+					   (atoi(p) == 2) ? _("Yes") : _("No"), "<BR>\n", NULL);
 			g_free(oldibuf);
 			break;
 		case 1:
@@ -782,6 +782,11 @@ static void import_buddies_server_results(struct gaim_connection *gc, gchar *web
 		gchar **data_tbl;
 		gchar *name, *show;
 
+		if (strlen(users_tbl[i])==0) {
+			debug_printf("import_buddies_server_results: users_tbl[i] is empty\n");
+			continue;
+		}
+
 		g_strdelimit(users_tbl[i], "\r\t\n\015", ' ');
 		data_tbl = g_strsplit(users_tbl[i], ";", 8);
 
@@ -793,7 +798,7 @@ static void import_buddies_server_results(struct gaim_connection *gc, gchar *web
 		}
 
 		debug_printf("import_buddies_server_results: uin: %s\n", name);
-		if (!find_buddy(gc->user, name)) {
+		if (!find_buddy(gc->account, name)) {
 			/* Default group if none specified on server */
 			gchar *group = g_strdup("Gadu-Gadu");
 			if (strlen(data_tbl[5])) {
@@ -805,7 +810,7 @@ static void import_buddies_server_results(struct gaim_connection *gc, gchar *web
 				g_strfreev(group_tbl);
 			}
 			/* Add Buddy to our userlist */
-			add_buddy(gc->user, group, name, strlen(show) ? show : name);
+			add_buddy(gc->account, group, name, strlen(show) ? show : name);
 			gaim_blist_save();
 			g_free(group);
 		}
@@ -844,7 +849,7 @@ static void password_change_server_results(struct gaim_connection *gc, gchar *we
 		return;
 	}
 
-	debug_printf("delete_buddies_server_results: webdata [%s]\n", webdata);
+	debug_printf("password_change_server_results: webdata [%s]\n", webdata);
 	do_error_dialog(_("Password couldn't be changed"), NULL, GAIM_ERROR);
 }
 
@@ -934,7 +939,7 @@ static void http_req_callback(gpointer data, gint source, GaimInputCondition con
 		return;
 	}
 
-	if (source != 0) {
+	if (source == 0) {
 		g_free(request);
 		g_free(hdata);
 		return;
@@ -1014,10 +1019,11 @@ static void export_buddies_server(struct gaim_connection *gc)
 	while (gr) {
 		struct group *g = gr->data;
 		GSList *m = g->members;
+		int num_added=0;
 		while (m) {
 			struct buddy *b = m->data;
 
-			if(b->user->gc == gc) {
+			if(b->account == gc->account) {
 				gchar *newdata;
 				/* GG Number */
 				gchar *name = gg_urlencode(b->name);
@@ -1027,9 +1033,15 @@ static void export_buddies_server(struct gaim_connection *gc)
 				gchar *gname = gg_urlencode(g->name);
 
 				ptr = he->request;
-				newdata = g_strdup_printf("%s;%s;%s;%s;%s;%s;%s\r\n",
+				newdata = g_strdup_printf("%s;%s;%s;%s;%s;%s;%s",
 						show, show, show, show, "", gname, name);
-				he->request = g_strconcat(ptr, newdata, NULL);
+
+				if(num_added > 0)
+					he->request = g_strconcat(ptr, "%0d%0a", newdata, NULL);
+				else
+					he->request = g_strconcat(ptr, newdata, NULL);
+
+				num_added++;
 
 				g_free(newdata);
 				g_free(ptr);
